#ifndef MYSTRING__H__
#define MYSTRING__H__

// Copyright (C) 1998 Bruce Guenter <bruce.guenter@qcc.sk.ca>
//
// This package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

#include <string.h>

struct mystringrep
{
  unsigned length;
  unsigned references;
  char buf[1];

  void attach();
  void detach();
  
  static mystringrep* alloc(unsigned);
  static mystringrep* dup(const char*, unsigned);
  static mystringrep* dup(const char*, unsigned,
			  const char*, unsigned);
};

extern mystringrep nil;

class mystring;
class mystringtmp
{
  friend class mystring;
private:
  mystringrep* rep;
  const char* str;
  unsigned len;
  static const unsigned badlen = (unsigned)-1;
  void checkstr();
protected:
  unsigned length();
  operator const char*();
  bool operator!();
  mystringrep* getrep();
public:
  mystringtmp(const mystringtmp&);
  mystringtmp(const char* s);
  mystringtmp(const char* s, unsigned l);
  mystringtmp(mystringrep* r);
  mystringtmp(const mystring&);
};
      
class mystring
{
  friend class mystringtmp;
private:
  mystringrep* rep;

protected:
  void dupnil();
  void dup(mystringtmp);
  void append(mystringtmp);
  void assign(mystringtmp);
public:
  mystring() { dupnil(); }
  mystring(const char* s);
  mystring(const mystring& s);
  mystring(const char* str, size_t len);
  mystring(mystringtmp a, mystringtmp b);
  mystring(mystringtmp a, mystringtmp b, mystringtmp c);
  mystring(mystringtmp a, mystringtmp b, mystringtmp c, mystringtmp d);
  mystring(mystringtmp a, mystringtmp b, mystringtmp c,
	   mystringtmp d, mystringtmp e);
  mystring(mystringtmp a, mystringtmp b, mystringtmp c,
	   mystringtmp d, mystringtmp e, mystringtmp f);
  ~mystring();

  const char* c_str() const { return rep->buf; }

  bool operator!() const { return empty(); }
  
  char operator[](size_t i) const { return rep->buf[i]; }
  
  size_t length() const { return rep->length; }

  bool empty() const { return rep->length == 0; }
  
  int operator!=(const char* in) const;
  int operator!=(const mystring& in) const;
  bool operator==(const char* in) const
    {
      return !operator!=(in);
    }
  bool operator==(const mystring& in) const
    {
      return !operator!=(in);
    }
  
  mystring operator=(const char* in);
  mystring operator=(const mystring& in);
  
  mystring subst(char from, char to) const;
  
  mystring lower() const;
  mystring upper() const;

  int find(char, size_t = 0) const;
  int find_first(mystringtmp, size_t = 0) const;

  int find_last(char) const;
  int find_last(mystringtmp) const;
  
  mystring left(size_t) const;
  mystring right(size_t) const;
  mystring sub(size_t, size_t) const;

  mystring lstrip() const;
  mystring rstrip() const;
  mystring strip() const;

  mystring operator+=(mystringtmp that);
  mystring operator+=(char ch);
};

mystring operator+(mystringtmp, mystringtmp);

typedef mystring string;

#endif
